// Linked list based bag after Sedgewick and Wayne, Algorithms 4th ed, Algorithm 1.2

#ifndef BAG_HPP
#define BAG_HPP

#include <cstddef>

template <typename T>
class Bag {

private:

////////////////////////////////////
    class Node {
    public:
        T item;
        Node *next;
        };


public:

    ////////////////////////////////////
    class iterator;
    friend class iterator;
    class iterator {

    public:

        iterator( Node *c );

        iterator& operator++();
        iterator operator++(int);

        T operator*();
        bool operator!=( typename Bag<T>::iterator );

    private:

        Node *curr;
        };

    ////////////////////////////////////

    Bag( void );
    ~Bag( void );

    void add( T &item );

    bool is_empty( void );
    size_t size( void );

    iterator begin( void );
    iterator end( void );

#ifdef EXERCISES
    T delete_bottom( void );
    T remove( size_t k );
    bool find( T key );
#endif

private:

    size_t N;
    Node *list;

    };


////////////////////////////////////////////////////////////////////////////////


template <typename T>
Bag<T>::Bag( void ) :
    N (0),
    list (nullptr)
    {
    }
    
template <typename T>
Bag<T>::~Bag( void ) {
    // This will leak memory, won't it?
    while ( list ) {
        Node * next = list->next;
        delete list;
        list = next;
        }
    }

template <typename T>
void Bag<T>::add( T &item ) {
    Node * node = new Node();
    node->item = item;
    node->next = list;
    list = node;
    N++;
    }

template <typename T>
bool Bag<T>::is_empty( void ) {
    return N == 0;
    }

template <typename T>
size_t Bag<T>::size( void ) {
    return N;
    }

template <typename T>
typename Bag<T>::iterator Bag<T>::begin( void ) {
    return Bag<T>::iterator( list );
    }

template <typename T>
typename Bag<T>::iterator Bag<T>::end( void ) {
    return Bag<T>::iterator( nullptr );
    }


////////////////////////////////////////////////////////////////////////////////

template <typename T>
Bag<T>::iterator::iterator( Node *c ) :
    curr (c)
    {
    }

template <typename T>
typename Bag<T>::iterator & Bag<T>::iterator::operator++() {
    if ( this->curr != nullptr ) {
        this->curr = this->curr->next;
        }
    return *this;
    }

template <typename T>
typename Bag<T>::iterator Bag<T>::iterator::operator++( int ) {
    auto t = Bag<T>::iterator( *this );

    if ( this->curr != nullptr ) {
        this->curr = this->curr->next;
        }
    return t;
    }

template <typename T>
T Bag<T>::iterator::operator*() {
    return this->curr->item;
    }

template <typename T>
bool Bag<T>::iterator::operator!=( typename Bag<T>::iterator other ) {
    return this->curr != other.curr;
    }

////////////////////////////////////////////////////////////////////////////////


#ifdef EXERCISES

// 1.3.19
// Returns the item if entry was found.  If the entry did not exist, then the returned value is undefined.
template <typename T>
T Bag<T>::delete_bottom( void ) {
    Node *prev, *curr;
    T item;

    prev = nullptr;
    curr = list;
    if ( curr ) {
        while ( curr->next ) {
            prev = curr;
            curr = curr->next;
            }

        item = curr->item;
        delete curr;
        if ( prev )
            prev->next = nullptr;
        else
            list = nullptr;

        N--;
        }

    return item;
    }


// 1.3.20
// Returns the item if entry was found.  If the entry did not exist, then the returned value is undefined.
template <typename T>
T Bag<T>::remove( size_t k ) {
    Node *prev, *curr;
    size_t i=0;
    T item;

    prev = nullptr;
    curr = list;
    while ( curr && (i < k) ) {
        prev = curr;
        curr = curr->next;
        ++i;
        }

    if ( curr &&( i == k) ) {
        item = curr->item;
        if ( prev )
            prev->next = curr->next;
        else
            list = curr->next;;
        delete curr;
        N--;
        }

    return item;
    }

// 1.3.21
// Returns true if the specified key is in the bag.
template <typename T>
bool Bag<T>::find( T key ) {
    Node *curr = list;

    while ( curr ) {
        if ( curr->item == key )
            return true;
        curr = curr->next;
        }

    return false;
    }

#endif


#endif

