#ifndef CIRCULARQUEUE_HPP

#define CIRCULARQUEUE_HPP

#include <cassert>
#include <cstddef>
#include <vector>

template <typename T>
class CircularQueue {
public:
    CircularQueue ( size_t max );

    void enqueue( T &item);
    T dequeue( void );

    bool empty( void );
    bool full( void );
    size_t size( void );

private:
    std::vector<T> data;
    size_t first;
    size_t end;
    size_t N;
    };


////////////////////////////////////////////////////////////////////////////////

template <typename T>
CircularQueue<T>::CircularQueue( size_t max ) :
    first (0),
    end (0),
    N (0)
    {
    data.resize( max );
    }

template <typename T>
void CircularQueue<T>::enqueue( T &item ) {
    data[end] = item;
    if ( N < data.size() )
        ++N;
    else {
        ++first;
        if ( first > data.size() )
            first = 0;
        }
    ++end;
    if ( end == data.size()  )
        end = 0;
    }

template <typename T>
T CircularQueue<T>::dequeue( void ) {
    assert( N > 0 );
    T item = data[first];
    if ( N > 0 )
        --N;
    ++first;
    if ( first == data.size() )
        first = 0;

    return item;
    }

template <typename T>
bool CircularQueue<T>::empty( void ) {
    return ( N == 0 );
    }

template <typename T>
bool CircularQueue<T>::full( void ) {
    return ( N == data.size() );
    }

template <typename T>
size_t CircularQueue<T>::size( void ) {
    return N;
    }



#endif
