// Linked list based deque after Sedgewick and Wayne, Algorithms 4th ed, Algorithm 1.3

#ifndef DEQUE_HPP
#define DEQUE_HPP

#include <cassert>
#include <cstddef>

template <typename T>
class Deque {

private:

    ////////////////////////////////////
    class Node {
    public:
        T item;
        Node *next;
        Node *prev;
        };

public:

    ////////////////////////////////////
    class iterator;
    friend class iterator;
    class iterator {

    public:

        iterator( Node *c );

        iterator& operator++();
        iterator operator++(int);

        T operator*();
        bool operator!=( typename Deque<T>::iterator );

    private:

        Node *curr;
        };

    ////////////////////////////////////

    Deque( void );
    ~Deque( void );

    void push_left( T &item );
    void push_right( T &item );
    T pop_left( void );
    T pop_right( void );

    bool is_empty( void );
    size_t size( void );

    iterator begin( void );
    iterator end( void );


private:

    size_t N;
    Node *first;
    Node *last;

    };


////////////////////////////////////////////////////////////////////////////////

template <typename T>
Deque<T>::Deque( void ) :
    N (0),
    first (nullptr),
    last (nullptr)
    {
    }
    
template <typename T>
Deque<T>::~Deque( void ) {
    assert( N == 0 );  // Catch the potential memory leak.
    while ( first ) {
        Node *next = first->next;
        delete first;
        first = next;
        }
    }

template <typename T>
void Deque<T>::push_right( T &item ) {
    Node * node = new Node();
    node->item = item;
    node->next = nullptr;
    node->prev = nullptr;

    if ( last ) {
        last->next = node;
        node->prev = last;
        last = node;
        }
    else {
        last = first = node;
        }

    N++;
    }

template <typename T>
void Deque<T>::push_left( T &item ) {
    Node * node = new Node();
    node->item = item;
    node->next = nullptr;
    node->prev = nullptr;

    if ( first ) {
        node->next = first;
        first->prev = node;
        first = node;
        }
    else {
        last = first = node;
        }

    N++;
    }

template <typename T>
T Deque<T>::pop_left( void ) {
    assert( N > 0 );

    T item = first->item;

    Node *old = first;
    if ( first->next )
        first->next->prev = nullptr;
    first = first->next;
    delete old;
    N--;

    if ( ! first )
        last = nullptr;

    return item;
    }

template <typename T>
T Deque<T>::pop_right( void ) {
    assert( N > 0 );

    T item = last->item;

    Node *old = last;
    if ( last->prev )
        last->prev->next = nullptr;
    last = last->prev;
    delete old;
    N--;

    if ( ! last )
        first = nullptr;

    return item;
    }

template <typename T>
bool Deque<T>::is_empty( void ) {
    return N == 0;
    }

template <typename T>
size_t Deque<T>::size( void ) {
    return N;
    }

template <typename T>
typename Deque<T>::iterator Deque<T>::begin( void ) {
    return Deque<T>::iterator( first );
    }

template <typename T>
typename Deque<T>::iterator Deque<T>::end( void ) {
    return Deque<T>::iterator( nullptr );
    }


////////////////////////////////////////////////////////////////////////////////

template <typename T>
Deque<T>::iterator::iterator( Node *c ) :
    curr (c)
    {
    }

template <typename T>
typename Deque<T>::iterator & Deque<T>::iterator::operator++() {
    if ( this->curr != nullptr ) {
        this->curr = this->curr->next;
        }
    return *this;
    }

template <typename T>
typename Deque<T>::iterator Deque<T>::iterator::operator++( int ) {
    auto t = Deque<T>::iterator( *this );

    if ( this->curr != nullptr ) {
        this->curr = this->curr->next;
        }
    return t;
    }

template <typename T>
T Deque<T>::iterator::operator*() {
    return this->curr->item;
    }

template <typename T>
bool Deque<T>::iterator::operator!=( typename Deque<T>::iterator other ) {
    return this->curr != other.curr;
    }

#endif

