// g++ -std=c++11 Queue.cpp


#include <iostream>
#include <typeinfo>

#include "Queue.hpp"

int main ( int argc, char **argv ) {

    Queue<long> queue;

    long i;
    while ( ! std::cin.eof() ) {
        std::cin >> i;
        if ( std::cin.good() ) 
            if ( i >= 0 )
                queue.enqueue(i);
            else
                queue.dequeue();
        }

    std::cout << "Queue has " << queue.size() << " entries." << std::endl;

    for ( auto iter = queue.begin(); iter != queue.end(); ++iter ) {
        std::cout << *iter << std::endl;
        }

    Queue<long> q2( queue );
    std::cout << "Copied queue to q2.  q2 has " << q2.size() << " entries." << std::endl;
    for ( auto iter = q2.begin(); iter != q2.end(); ++iter ) {
        std::cout << *iter << std::endl;
        }

    // Just make sure that we can build an iterator with the copy constructor.
    auto iter = q2.begin();
    auto iter2( iter );

    std::cout << "Dequeuing entries..." << std::endl;

    while ( ! queue.is_empty() ) {
        i = queue.dequeue();
        std::cout << i << std::endl;
        }

    std::cout << "Queue has " << queue.size() << " entries." << std::endl;


    // Silently empty q2 to avoid the assertion in the destructor.
    while ( ! q2.is_empty() )
        q2.dequeue();

    // Show the iterator traits;
    std::cout << std::endl;
    if ( typeid( Queue<long>::iterator_traits::iterator_category) == typeid( std::forward_iterator_tag ) )
        std::cout << "category: forward iterator" << std::endl;;
    if ( typeid( Queue<long>::iterator_traits::value_type) == typeid( long ) )
        std::cout << "value_type: long" << std::endl;;
    if ( typeid( Queue<long>::iterator_traits::pointer) == typeid( long * ) )
        std::cout << "pointer: long *" << std::endl;;
    if ( typeid( Queue<long>::iterator_traits::reference) == typeid( long & ) )
        std::cout << "reference: long &" << std::endl;;
    if ( typeid( Queue<long>::iterator_traits::difference_type ) == typeid( ptrdiff_t ) )
        std::cout << "difference_type: ptrdiff_t" << std::endl;;
    }

