// Bag class with random iteration order.

#ifndef RANDOMBAG_HPP
#define RANDOMBAG_HPP

#include <algorithm>
#include <cstddef>
#include <chrono>
#include <iostream>
#include <iterator>
#include <random> 
#include <vector>

template <typename T>
class RandomBag {

public:

    ////////////////////////////////////////
    class iterator {

    public:

        iterator( T *b, T *f, T *e );

        iterator& operator++();
        iterator operator++(int);

        T operator*();
        bool operator!=( typename RandomBag<T>::iterator other );

    private:

        T *begin;
        T *first;
        T *end;
        size_t curr;
        std::vector<size_t> order;
        };

    ////////////////////////////////////////
    RandomBag( void );
    ~RandomBag( void );

    void add( T &item );

    bool is_empty( void );
    size_t size( void );

    iterator begin( void );
    iterator end( void );

private:

    size_t N;
    size_t max;
    T *data;

    void resize( size_t new_max );
    };


////////////////////////////////////////////////////////////////////////////////

template <typename T>
RandomBag<T>::RandomBag( void ) :
    N (0),
    max(1) {
    data = new T[1];
    }
    
template <typename T>
RandomBag<T>::~RandomBag( void ) {
    // May leak memory.
    if ( data )
        delete[] data;
    }

template <typename T>
void RandomBag<T>::add( T &item ) {
    if ( max == N )
        resize( 2 * max );
    data[N++] = item;
    }

template <typename T>
bool RandomBag<T>::is_empty( void ) {
    return N == 0;
    }

template <typename T>
size_t RandomBag<T>::size( void ) {
    return N;
    }

template <typename T>
void RandomBag<T>::resize( size_t new_max ) {
    T *new_data = new T[new_max];
    for ( size_t i = 0; i < N; ++i )
        new_data[i] = data[i];
    T *old_data = data;
    data = new_data;
    delete[] old_data;
    max = new_max;
    }

template <typename T>
typename RandomBag<T>::iterator RandomBag<T>::begin( void ) {
    return RandomBag<T>::iterator( data, data, data+N );
    }

template <typename T>
typename RandomBag<T>::iterator RandomBag<T>::end( void ) {
    return RandomBag<T>::iterator( data, data+N, data+N );
    }


////////////////////////////////////////////////////////////////////////////////

template <typename T>
RandomBag<T>::iterator::iterator( T *b, T *f, T *e ) :
    begin (b),
    first (f),
    end (e),
    curr (0) {
    size_t size = e - f;
    for ( size_t i = 0; i < size; ++i ) {
        order.push_back( (first - begin) + i );
        }

    unsigned seed = std::chrono::system_clock::now().time_since_epoch().count();
    shuffle (order.begin(), order.end(), std::default_random_engine(seed));

    order.push_back( end - begin );
    }

template <typename T>
typename RandomBag<T>::iterator & RandomBag<T>::iterator::operator++() {
    if ( this->curr < this->end - this->first ) {
        this->curr += 1;
        }
    return *this;
    }

template <typename T>
typename RandomBag<T>::iterator RandomBag<T>::iterator::operator++( int ) {
    auto t = RandomBag<T>::iterator( *this );
    if ( this->curr < this->end - this->first ) {
        this->curr += 1;
        }
    return t;
    }

template <typename T>
T RandomBag<T>::iterator::operator*() {
    return *(this->begin + order[curr]);
    }

template <typename T>
bool RandomBag<T>::iterator::operator!=( typename RandomBag<T>::iterator other ) {
    return (this->begin + this->order[this->curr]) != (other.begin + other.order[other.curr]);
    }



#endif

