// Stack built on a resizing array.

#ifndef RESIZINGARRAYSTACK_HPP
#define RESIZINGARRAYSTACK_HPP

#include <cassert>
#include <cstddef>

template <typename T>
class ResizingArrayStack {

public:

    ResizingArrayStack( void );
    ~ResizingArrayStack( void );

    void push( T &item );
    T pop( void );

    bool is_empty( void );
    size_t size( void );

    class iterator;
    friend class iterator;
    class iterator {

    public:

        iterator( T *c, T *e );

        iterator& operator++();
        iterator operator++(int);

        T operator*();
        bool operator!=( typename ResizingArrayStack<T>::iterator );

    private:

        T *curr;
        T *end;
        };

    iterator begin( void );
    iterator end( void );

private:

    size_t N;
    size_t max;
    T *data;

    void resize( size_t new_max );
    };

////////////////////////////////////////////////////////////////////////////////

template <typename T>
ResizingArrayStack<T>::ResizingArrayStack( void ) :
    N (0),
    max(1) {
    data = new T[1];
    }
    
template <typename T>
ResizingArrayStack<T>::~ResizingArrayStack( void ) {
    assert( N == 0 );  // Catch the potential memory leak.
    if ( data )
        delete[] data;
    }

template <typename T>
void ResizingArrayStack<T>::push( T &item ) {
    if ( max == N )
        resize( 2 * max );
    data[N++] = item;
    }

template <typename T>
T ResizingArrayStack<T>::pop( void ) {
    assert( N > 0 );
    T item = data[N-1];
    N--;

    if ( N <= max/4 )
        resize( max/2 );

    return item;
    }

template <typename T>
bool ResizingArrayStack<T>::is_empty( void ) {
    return N == 0;
    }

template <typename T>
size_t ResizingArrayStack<T>::size( void ) {
    return N;
    }

template <typename T>
void ResizingArrayStack<T>::resize( size_t new_max ) {
    T *new_data = new T[new_max];
    for ( size_t i = 0; i < N; ++i )
        new_data[i] = data[i];
    T *old_data = data;
    data = new_data;
    delete[] old_data;
    max = new_max;
    }

template <typename T>
typename ResizingArrayStack<T>::iterator ResizingArrayStack<T>::begin( void ) {
    return ResizingArrayStack<T>::iterator( data+N-1, data-1 );
    }

template <typename T>
typename ResizingArrayStack<T>::iterator ResizingArrayStack<T>::end( void ) {
    return ResizingArrayStack<T>::iterator( data-1, data-1 );
    }


////////////////////////////////////////////////////////////////////////////////

template <typename T>
ResizingArrayStack<T>::iterator::iterator( T *c, T *e ) :
    curr (c),
    end (e)
    {
    }

template <typename T>
typename ResizingArrayStack<T>::iterator & ResizingArrayStack<T>::iterator::operator++() {
    if ( this->curr > this->end ) {
        this->curr -= 1;
        }
    return *this;
    }

template <typename T>
typename ResizingArrayStack<T>::iterator ResizingArrayStack<T>::iterator::operator++( int ) {
    auto t = ResizingArrayStack<T>::iterator( *this );
    if ( this->curr > this->end ) {
        this->curr -= 1;
        }
    return t;
    }

template <typename T>
T ResizingArrayStack<T>::iterator::operator*() {
    return *(this->curr);
    }

template <typename T>
bool ResizingArrayStack<T>::iterator::operator!=( typename ResizingArrayStack<T>::iterator other ) {
    return this->curr != other.curr;
    }



#endif

