// Linked list based stack after Sedgewick and Wayne, Algorithms 4th ed, Algorithm 1.2

#ifndef STACK_HPP
#define STACK_HPP

#include <cassert>
#include <cstddef>


template <typename T>
class Stack {

private:

    ////////////////////////////////////
    class Node {
    public:
        T item;
        Node *next;
        };


public:

    ////////////////////////////////////
    class iterator;
    friend class iterator;
    class iterator {

    public:

        iterator( Node *c );

        iterator& operator++();
        iterator operator++(int);

        T operator*();
        bool operator!=( typename Stack<T>::iterator );

    private:

        Node *curr;
        };

    ////////////////////////////////////

    Stack( void );
    ~Stack( void );

    void push( T &item );
    T pop( void );

    bool is_empty( void );
    size_t size( void );

    iterator begin( void );
    iterator end( void );

#ifdef EXERCISES

    T delete_bottom( void );    // 1.3.19
    T remove( size_t k );       // 1.3.20
    bool find( T key );         // 1.3.21

#endif

private:

    size_t N;
    Node *list;

    };


////////////////////////////////////////////////////////////////////////////////


template <typename T>
Stack<T>::Stack( void ) :
    N (0),
    list (nullptr)
    {
    }
    
template <typename T>
Stack<T>::~Stack( void ) {
    assert( N == 0 );  // Catch the potential memory leak.
    while ( list ) {
        Node * next = list->next;
        delete list;
        list = next;
        }
    }

template <typename T>
void Stack<T>::push( T &item ) {
    Node * node = new Node();
    node->item = item;
    node->next = list;
    list = node;
    N++;
    }

template <typename T>
T Stack<T>::pop( void ) {
    assert( N > 0 );

    T item = list->item;

    Node *old = list;
    list = list->next;
    delete old;
    N--;

    return item;
    }

template <typename T>
bool Stack<T>::is_empty( void ) {
    return N == 0;
    }

template <typename T>
size_t Stack<T>::size( void ) {
    return N;
    }

template <typename T>
typename Stack<T>::iterator Stack<T>::begin( void ) {
    return Stack<T>::iterator( list );
    }

template <typename T>
typename Stack<T>::iterator Stack<T>::end( void ) {
    return Stack<T>::iterator( nullptr );
    }


////////////////////////////////////////////////////////////////////////////////

template <typename T>
Stack<T>::iterator::iterator( Node *c ) :
    curr (c)
    {
    }

template <typename T>
typename Stack<T>::iterator & Stack<T>::iterator::operator++() {
    if ( this->curr != nullptr ) {
        this->curr = this->curr->next;
        }
    return *this;
    }

template <typename T>
typename Stack<T>::iterator Stack<T>::iterator::operator++( int ) {
    auto t = Stack<T>::iterator( *this );

    if ( this->curr != nullptr ) {
        this->curr = this->curr->next;
        }
    return t;
    }

template <typename T>
T Stack<T>::iterator::operator*() {
    return this->curr->item;
    }

template <typename T>
bool Stack<T>::iterator::operator!=( typename Stack<T>::iterator other ) {
    return this->curr != other.curr;
    }

////////////////////////////////////////////////////////////////////////////////


#ifdef EXERCISES

// 1.3.19
// Returns the item if entry was found.  If the entry did not exist, then the returned value is undefined.
template <typename T>
T Stack<T>::delete_bottom( void ) {
    Node *prev, *curr;
    T item;

    prev = nullptr;
    curr = list;
    if ( curr ) {
        while ( curr->next ) {
            prev = curr;
            curr = curr->next;
            }

        item = curr->item;
        delete curr;
        if ( prev )
            prev->next = nullptr;
        else
            list = nullptr;

        N--;
        }

    return item;
    }


// 1.3.20
// Returns the item if entry was found.  If the entry did not exist, then the returned value is undefined.
template <typename T>
T Stack<T>::remove( size_t k ) {
    Node *prev, *curr;
    size_t i=0;
    T item;

    prev = nullptr;
    curr = list;
    while ( curr && (i < k) ) {
        prev = curr;
        curr = curr->next;
        ++i;
        }

    if ( curr &&( i == k) ) {
        item = curr->item;
        if ( prev )
            prev->next = curr->next;
        else
            list = curr->next;;
        delete curr;
        N--;
        }

    return item;
    }

// 1.3.21
// Returns true if the specified key is in the stack.
template <typename T>
bool Stack<T>::find( T key ) {
    Node *curr = list;

    while ( curr ) {
        if ( curr->item == key )
            return true;
        curr = curr->next;
        }

    return false;
    }

#endif


#endif

