// Linked list based steque after Sedgewick and Wayne, Algorithms 4th ed, Algorithm 1.2

#ifndef STEQUE_HPP
#define STEQUE_HPP

#include <cassert>
#include <cstddef>

template <typename T>
class Steque {

private:

    ////////////////////////////////////
    class Node {
    public:
        T item;
        Node *next;
        };


public:

    ////////////////////////////////////
    class iterator;
    friend class iterator;
    class iterator {

    public:

        iterator( Node *c );

        iterator& operator++();
        iterator operator++(int);

        T operator*();
        bool operator!=( typename Steque<T>::iterator );

    private:

        Node *curr;
        };

    ////////////////////////////////////

    Steque( void );
    ~Steque( void );

    void push( T &item );
    T pop( void );
    void enqueue( T &item );

    bool is_empty( void );
    size_t size( void );

    iterator begin( void );
    iterator end( void );

private:

    size_t N;
    Node *list;

    };


////////////////////////////////////////////////////////////////////////////////

template <typename T>
Steque<T>::Steque( void ) :
    N (0),
    list (nullptr)
    {
    }
    
template <typename T>
Steque<T>::~Steque( void ) {
    assert( N == 0 );  // Catch the potential memory leak.
    while ( list ) {
        Node * next = list->next;
        delete list;
        list = next;
        }
    }

template <typename T>
void Steque<T>::push( T &item ) {
    Node * node = new Node();
    node->item = item;
    node->next = list;
    list = node;
    N++;
    }

template <typename T>
T Steque<T>::pop( void ) {
    assert( N > 0 );

    T item = list->item;

    Node *old = list;
    list = list->next;
    delete old;
    N--;

    return item;
    }

template <typename T>
void Steque<T>::enqueue( T &item ) {
    Node *curr = list;

    Node *node = new Node();
    node->item = item;
    node->next = nullptr;

    if ( ! curr ) {
        list = node;
        N++;
        return;
        }

    while ( curr->next ) {
        curr = curr->next;
        }

    curr->next = node;
    N++;

    return;
    }

template <typename T>
bool Steque<T>::is_empty( void ) {
    return N == 0;
    }

template <typename T>
size_t Steque<T>::size( void ) {
    return N;
    }

template <typename T>
typename Steque<T>::iterator Steque<T>::begin( void ) {
    return Steque<T>::iterator( list );
    }

template <typename T>
typename Steque<T>::iterator Steque<T>::end( void ) {
    return Steque<T>::iterator( nullptr );
    }


////////////////////////////////////////////////////////////////////////////////

template <typename T>
Steque<T>::iterator::iterator( Node *c ) :
    curr (c)
    {
    }

template <typename T>
typename Steque<T>::iterator & Steque<T>::iterator::operator++() {
    if ( this->curr != nullptr ) {
        this->curr = this->curr->next;
        }
    return *this;
    }

template <typename T>
typename Steque<T>::iterator Steque<T>::iterator::operator++( int ) {
    auto t = Steque<T>::iterator( *this );

    if ( this->curr != nullptr ) {
        this->curr = this->curr->next;
        }
    return t;
    }

template <typename T>
T Steque<T>::iterator::operator*() {
    return this->curr->item;
    }

template <typename T>
bool Steque<T>::iterator::operator!=( typename Steque<T>::iterator other ) {
    return this->curr != other.curr;
    }


#endif

